@@ -9,6 +9,7 @@
  * Copyright (C) 2011 Renata Hodovan <reni@webkit.org>
  * Copyright (C) 2011 University of Szeged
  * Copyright (C) 2018 Adobe Systems Incorporated. All rights reserved.
+ * Copyright (C) 2020, 2021, 2022 Igalia S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -29,6 +30,7 @@
 #include "config.h"
 #include "RenderSVGShape.h"
 
+#if ENABLE(LAYER_BASED_SVG_ENGINE)
 #include "FloatPoint.h"
 #include "FloatQuad.h"
 #include "GraphicsContext.h"
@@ -38,12 +40,12 @@
 #include "PointerEventsHitRules.h"
 #include "RenderSVGResourceMarker.h"
 #include "RenderSVGResourceSolidColor.h"
+#include "RenderSVGShapeInlines.h"
 #include "SVGPathData.h"
 #include "SVGRenderingContext.h"
 #include "SVGResources.h"
 #include "SVGResourcesCache.h"
 #include "SVGURIReference.h"
-#include "StrokeStyleApplier.h"
 #include <wtf/IsoMallocInlines.h>
 #include <wtf/StackStats.h>
 
@@ -51,27 +53,8 @@ namespace WebCore {
 
 WTF_MAKE_ISO_ALLOCATED_IMPL(RenderSVGShape);
 
-class BoundingRectStrokeStyleApplier final : public StrokeStyleApplier {
-public:
-    BoundingRectStrokeStyleApplier(const RenderSVGShape& renderer)
-        : m_renderer(renderer)
-    {
-    }
-
-    void strokeStyle(GraphicsContext* context) override
-    {
-        SVGRenderSupport::applyStrokeStyleToContext(context, m_renderer.style(), m_renderer);
-    }
-
-private:
-    const RenderSVGShape& m_renderer;
-};
-
 RenderSVGShape::RenderSVGShape(SVGGraphicsElement& element, RenderStyle&& style)
     : RenderSVGModelObject(element, WTFMove(style))
-    , m_needsBoundariesUpdate(false) // Default is false, the cached rects are empty from the beginning.
-    , m_needsShapeUpdate(true) // Default is true, so we grab a Path object once from SVGGraphicsElement.
-    , m_needsTransformUpdate(true) // Default is true, so we grab a AffineTransform object once from SVGGraphicsElement.
 {
 }
 
@@ -114,16 +97,18 @@ void RenderSVGShape::strokeShape(GraphicsContext& context) const
 bool RenderSVGShape::shapeDependentStrokeContains(const FloatPoint& point, PointCoordinateSpace pointCoordinateSpace)
 {
     ASSERT(m_path);
-    BoundingRectStrokeStyleApplier applier(*this);
 
     if (hasNonScalingStroke() && pointCoordinateSpace != LocalCoordinateSpace) {
         AffineTransform nonScalingTransform = nonScalingStrokeTransform();
         Path* usePath = nonScalingStrokePath(m_path.get(), nonScalingTransform);
-
-        return usePath->strokeContains(&applier, nonScalingTransform.mapPoint(point));
+        return usePath->strokeContains(nonScalingTransform.mapPoint(point), [this] (GraphicsContext& context) {
+            SVGRenderSupport::applyStrokeStyleToContext(context, style(), *this);
+        });
     }
 
-    return m_path->strokeContains(&applier, point);
+    return m_path->strokeContains(point, [this] (GraphicsContext& context) {
+        SVGRenderSupport::applyStrokeStyleToContext(context, style(), *this);
+    });
 }
 
 bool RenderSVGShape::shapeDependentFillContains(const FloatPoint& point, const WindRule fillRule) const
@@ -158,32 +143,23 @@ bool RenderSVGShape::strokeContains(const FloatPoint& point, bool requiresStroke
 void RenderSVGShape::layout()
 {
     StackStats::LayoutCheckPoint layoutCheckPoint;
-    LayoutRepainter repainter(*this, SVGRenderSupport::checkForSVGRepaintDuringLayout(*this) && selfNeedsLayout());
 
-    bool updateCachedBoundariesInParents = false;
-
-    if (m_needsShapeUpdate || m_needsBoundariesUpdate) {
+    LayoutRepainter repainter(*this, checkForRepaintDuringLayout(), RepaintOutlineBounds::No);
+    if (m_needsShapeUpdate) {
+        // FIXME: [LBSE] Upstream SVGLengthContext changes
+        // graphicsElement().updateLengthContext();
         updateShapeFromElement();
+
         m_needsShapeUpdate = false;
-        updateRepaintBoundingBox();
-        m_needsBoundariesUpdate = false;
-        updateCachedBoundariesInParents = true;
+        setCurrentSVGLayoutRect(enclosingLayoutRect(m_fillBoundingBox));
     }
 
-    if (m_needsTransformUpdate) {
-        m_localTransform = graphicsElement().animatedLocalTransform();
-        m_needsTransformUpdate = false;
-        updateCachedBoundariesInParents = true;
-    }
+    updateLayerTransform();
 
     // Invalidate all resources of this client if our layout changed.
     if (everHadLayout() && selfNeedsLayout())
         SVGResourcesCache::clientLayoutChanged(*this);
 
-    // If our bounds changed, notify the parents.
-    if (updateCachedBoundariesInParents)
-        RenderSVGModelObject::setNeedsBoundariesUpdate();
-
     repainter.repaintAfterLayout();
     clearNeedsLayout();
 }
@@ -200,7 +176,7 @@ Path* RenderSVGShape::nonScalingStrokePath(const Path* path, const AffineTransfo
 
 bool RenderSVGShape::setupNonScalingStrokeContext(AffineTransform& strokeTransform, GraphicsContextStateSaver& stateSaver)
 {
-    Optional<AffineTransform> inverse = strokeTransform.inverse();
+    std::optional<AffineTransform> inverse = strokeTransform.inverse();
     if (!inverse)
         return false;
 
@@ -294,42 +270,46 @@ void RenderSVGShape::fillStrokeMarkers(PaintInfo& childPaintInfo)
     }
 }
 
-void RenderSVGShape::paint(PaintInfo& paintInfo, const LayoutPoint&)
+void RenderSVGShape::paint(PaintInfo& paintInfo, const LayoutPoint& paintOffset)
 {
-    if (paintInfo.context().paintingDisabled() || paintInfo.phase != PaintPhase::Foreground
-        || style().visibility() == Visibility::Hidden || isEmpty())
-        return;
-    FloatRect boundingBox = repaintRectInLocalCoordinates();
-    if (!SVGRenderSupport::paintInfoIntersectsRepaintRect(boundingBox, m_localTransform, paintInfo))
+    OptionSet<PaintPhase> relevantPaintPhases { PaintPhase::Foreground, PaintPhase::ClippingMask, PaintPhase::Mask, PaintPhase::Outline, PaintPhase::SelfOutline };
+    if (!shouldPaintSVGRenderer(paintInfo, relevantPaintPhases) || isEmpty())
         return;
 
-    PaintInfo childPaintInfo(paintInfo);
-    GraphicsContextStateSaver stateSaver(childPaintInfo.context());
-    childPaintInfo.applyTransform(m_localTransform);
-
-    if (childPaintInfo.phase == PaintPhase::Foreground) {
-        SVGRenderingContext renderingContext(*this, childPaintInfo);
-
-        if (renderingContext.isRenderingPrepared()) {
-            const SVGRenderStyle& svgStyle = style().svgStyle();
-            if (svgStyle.shapeRendering() == ShapeRendering::CrispEdges)
-                childPaintInfo.context().setShouldAntialias(false);
-
-            fillStrokeMarkers(childPaintInfo);
-        }
+    if (paintInfo.phase == PaintPhase::ClippingMask) {
+        // FIXME: [LBSE] Upstream SVGRenderSupport changes
+        // SVGRenderSupport::paintSVGClippingMask(*this, paintInfo);
+        return;
     }
 
-    if (style().outlineWidth())
-        paintOutline(childPaintInfo, IntRect(boundingBox));
-}
+    auto adjustedPaintOffset = paintOffset + currentSVGLayoutLocation();
+    if (paintInfo.phase == PaintPhase::Mask) {
+        // FIXME: [LBSE] Upstream SVGRenderSupport changes
+        // SVGRenderSupport::paintSVGMask(*this, paintInfo, adjustedPaintOffset);
+        return;
+    }
 
-// This method is called from inside paintOutline() since we call paintOutline()
-// while transformed to our coord system, return local coords
-void RenderSVGShape::addFocusRingRects(Vector<LayoutRect>& rects, const LayoutPoint&, const RenderLayerModelObject*)
-{
-    LayoutRect rect = LayoutRect(repaintRectInLocalCoordinates());
-    if (!rect.isEmpty())
-        rects.append(rect);
+    auto visualOverflowRect = visualOverflowRectEquivalent();
+    visualOverflowRect.moveBy(adjustedPaintOffset);
+    if (!visualOverflowRect.intersects(paintInfo.rect))
+        return;
+
+    if (paintInfo.phase == PaintPhase::Outline || paintInfo.phase == PaintPhase::SelfOutline) {
+        // FIXME: [LBSE] Upstream outline painting
+        // paintSVGOutline(paintInfo, adjustedPaintOffset);
+        return;
+    }
+
+    ASSERT(paintInfo.phase == PaintPhase::Foreground);
+    GraphicsContextStateSaver stateSaver(paintInfo.context());
+
+    auto coordinateSystemOriginTranslation = adjustedPaintOffset - nominalSVGLayoutLocation();
+    paintInfo.context().translate(coordinateSystemOriginTranslation.width(), coordinateSystemOriginTranslation.height());
+
+    if (style().svgStyle().shapeRendering() == ShapeRendering::CrispEdges)
+        paintInfo.context().setShouldAntialias(false);
+
+    fillStrokeMarkers(paintInfo);
 }
 
 bool RenderSVGShape::isPointInFill(const FloatPoint& point)
@@ -355,32 +335,46 @@ FloatPoint RenderSVGShape::getPointAtLength(float distance) const
     return hasPath() ? path().pointAtLength(distance) : createPath()->pointAtLength(distance);
 }
 
-bool RenderSVGShape::nodeAtFloatPoint(const HitTestRequest& request, HitTestResult& result, const FloatPoint& pointInParent, HitTestAction hitTestAction)
+bool RenderSVGShape::nodeAtPoint(const HitTestRequest& request, HitTestResult& result, const HitTestLocation& locationInContainer, const LayoutPoint& accumulatedOffset, HitTestAction hitTestAction)
 {
-    // We only draw in the forground phase, so we only hit-test then.
     if (hitTestAction != HitTestForeground)
         return false;
 
-    FloatPoint localPoint = m_localTransform.inverse().valueOr(AffineTransform()).mapPoint(pointInParent);
+    auto adjustedLocation = accumulatedOffset + currentSVGLayoutLocation();
+
+    auto localPoint = locationInContainer.point();
+    auto coordinateSystemOriginTranslation = nominalSVGLayoutLocation() - adjustedLocation;
+    localPoint.move(coordinateSystemOriginTranslation);
 
     if (!SVGRenderSupport::pointInClippingArea(*this, localPoint))
         return false;
 
-    PointerEventsHitRules hitRules(PointerEventsHitRules::SVG_PATH_HITTESTING, request, style().pointerEvents());
+    SVGHitTestCycleDetectionScope hitTestScope(*this);
+
+    PointerEventsHitRules hitRules(PointerEventsHitRules::SVG_PATH_HITTESTING, request, style().effectivePointerEvents());
     bool isVisible = (style().visibility() == Visibility::Visible);
     if (isVisible || !hitRules.requireVisible) {
         const SVGRenderStyle& svgStyle = style().svgStyle();
         WindRule fillRule = svgStyle.fillRule();
         if (request.svgClipContent())
             fillRule = svgStyle.clipRule();
-        if ((hitRules.canHitStroke && (svgStyle.hasStroke() || !hitRules.requireStroke) && strokeContains(localPoint, hitRules.requireStroke))
-            || (hitRules.canHitFill && (svgStyle.hasFill() || !hitRules.requireFill) && fillContains(localPoint, hitRules.requireFill, fillRule))
-            || (hitRules.canHitBoundingBox && objectBoundingBox().contains(localPoint))) {
-            updateHitTestResult(result, LayoutPoint(localPoint));
-            if (result.addNodeToListBasedTestResult(&graphicsElement(), request, localPoint) == HitTestProgress::Stop)
+
+        if (hitRules.canHitStroke && (svgStyle.hasStroke() || !hitRules.requireStroke) && strokeContains(localPoint, hitRules.requireStroke)) {
+            updateHitTestResult(result, locationInContainer.point() - toLayoutSize(adjustedLocation));
+            if (result.addNodeToListBasedTestResult(nodeForHitTest(), request, locationInContainer, m_strokeBoundingBox) == HitTestProgress::Stop)
                 return true;
+            return false;
+        }
+
+        if ((hitRules.canHitFill && (svgStyle.hasFill() || !hitRules.requireFill) && fillContains(localPoint, hitRules.requireFill, fillRule))
+            || (hitRules.canHitBoundingBox && m_fillBoundingBox.contains(localPoint))) {
+            updateHitTestResult(result, locationInContainer.point() - toLayoutSize(adjustedLocation));
+            if (result.addNodeToListBasedTestResult(nodeForHitTest(), request, locationInContainer, m_fillBoundingBox) == HitTestProgress::Stop)
+                return true;
+            return false;
         }
     }
+
     return false;
 }
 
@@ -399,23 +393,28 @@ static inline RenderSVGResourceMarker* markerForType(SVGMarkerType type, RenderS
     return 0;
 }
 
-FloatRect RenderSVGShape::markerRect(float strokeWidth) const
+FloatRect RenderSVGShape::computeMarkerBoundingBox(const SVGBoundingBoxComputation::DecorationOptions&) const
 {
-    ASSERT(!m_markerPositions.isEmpty());
+    if (m_markerPositions.isEmpty())
+        return FloatRect();
 
     auto* resources = SVGResourcesCache::cachedResourcesForRenderer(*this);
     ASSERT(resources);
 
-    RenderSVGResourceMarker* markerStart = resources->markerStart();
-    RenderSVGResourceMarker* markerMid = resources->markerMid();
-    RenderSVGResourceMarker* markerEnd = resources->markerEnd();
-    ASSERT(markerStart || markerMid || markerEnd);
+    auto* markerStart = resources->markerStart();
+    auto* markerMid = resources->markerMid();
+    auto* markerEnd = resources->markerEnd();
+    if (!markerStart && !markerMid && !markerEnd)
+        return FloatRect();
 
     FloatRect boundaries;
     unsigned size = m_markerPositions.size();
     for (unsigned i = 0; i < size; ++i) {
-        if (RenderSVGResourceMarker* marker = markerForType(m_markerPositions[i].type, markerStart, markerMid, markerEnd))
-            boundaries.unite(marker->markerBoundaries(marker->markerTransformation(m_markerPositions[i].origin, m_markerPositions[i].angle, strokeWidth)));
+        if (auto* marker = markerForType(m_markerPositions[i].type, markerStart, markerMid, markerEnd)) {
+            // FIXME: [LBSE] Upstream RenderSVGResourceMarker changes
+            // boundaries.unite(marker->computeMarkerBoundingBox(options, marker->markerTransformation(m_markerPositions[i].origin, m_markerPositions[i].angle, strokeWidth())));
+            boundaries.unite(marker->markerBoundaries(marker->markerTransformation(m_markerPositions[i].origin, m_markerPositions[i].angle, strokeWidth())));
+        }
     }
     return boundaries;
 }
@@ -430,35 +429,27 @@ FloatRect RenderSVGShape::calculateStrokeBoundingBox() const
     ASSERT(m_path);
     FloatRect strokeBoundingBox = m_fillBoundingBox;
 
-    const SVGRenderStyle& svgStyle = style().svgStyle();
-    if (svgStyle.hasStroke()) {
-        BoundingRectStrokeStyleApplier strokeStyle(*this);
+    if (style().svgStyle().hasStroke()) {
         if (hasNonScalingStroke()) {
             AffineTransform nonScalingTransform = nonScalingStrokeTransform();
-            if (Optional<AffineTransform> inverse = nonScalingTransform.inverse()) {
+            if (std::optional<AffineTransform> inverse = nonScalingTransform.inverse()) {
                 Path* usePath = nonScalingStrokePath(m_path.get(), nonScalingTransform);
-                FloatRect strokeBoundingRect = usePath->strokeBoundingRect(&strokeStyle);
+                FloatRect strokeBoundingRect = usePath->strokeBoundingRect(Function<void(GraphicsContext&)> { [this] (GraphicsContext& context) {
+                    SVGRenderSupport::applyStrokeStyleToContext(context, style(), *this);
+                } });
                 strokeBoundingRect = inverse.value().mapRect(strokeBoundingRect);
                 strokeBoundingBox.unite(strokeBoundingRect);
             }
-        } else
-            strokeBoundingBox.unite(path().strokeBoundingRect(&strokeStyle));
+        } else {
+            strokeBoundingBox.unite(path().strokeBoundingRect(Function<void(GraphicsContext&)> { [this] (GraphicsContext& context) {
+                SVGRenderSupport::applyStrokeStyleToContext(context, style(), *this);
+            } }));
+        }
     }
 
-    if (!m_markerPositions.isEmpty())
-        strokeBoundingBox.unite(markerRect(strokeWidth()));
-
     return strokeBoundingBox;
 }
 
-void RenderSVGShape::updateRepaintBoundingBox()
-{
-    m_repaintBoundingBoxExcludingShadow = strokeBoundingBox();
-    SVGRenderSupport::intersectRepaintRectWithResources(*this, m_repaintBoundingBoxExcludingShadow);
-
-    m_repaintBoundingBox = m_repaintBoundingBoxExcludingShadow;
-}
-
 float RenderSVGShape::strokeWidth() const
 {
     SVGLengthContext lengthContext(&graphicsElement());
@@ -474,7 +465,7 @@ bool RenderSVGShape::hasSmoothStroke() const
         && style().capStyle() == style().initialCapStyle();
 }
 
-void RenderSVGShape::drawMarkers(PaintInfo& paintInfo)
+void RenderSVGShape::drawMarkers(PaintInfo&)
 {
     ASSERT(!m_markerPositions.isEmpty());
 
@@ -482,17 +473,25 @@ void RenderSVGShape::drawMarkers(PaintInfo& paintInfo)
     if (!resources)
         return;
 
-    RenderSVGResourceMarker* markerStart = resources->markerStart();
-    RenderSVGResourceMarker* markerMid = resources->markerMid();
-    RenderSVGResourceMarker* markerEnd = resources->markerEnd();
+    auto* markerStart = resources->markerStart();
+    auto* markerMid = resources->markerMid();
+    auto* markerEnd = resources->markerEnd();
     if (!markerStart && !markerMid && !markerEnd)
         return;
 
     float strokeWidth = this->strokeWidth();
     unsigned size = m_markerPositions.size();
     for (unsigned i = 0; i < size; ++i) {
-        if (RenderSVGResourceMarker* marker = markerForType(m_markerPositions[i].type, markerStart, markerMid, markerEnd))
-            marker->draw(paintInfo, marker->markerTransformation(m_markerPositions[i].origin, m_markerPositions[i].angle, strokeWidth));
+        if (auto* marker = markerForType(m_markerPositions[i].type, markerStart, markerMid, markerEnd)) {
+            UNUSED_PARAM(marker);
+            UNUSED_PARAM(strokeWidth);
+
+            // FIXME: [LBSE] Upstream RenderLayer changes
+            // ASSERT(marker->hasLayer());
+            // GraphicsContextStateSaver stateSaver(paintInfo.context());
+            // auto contentTransform = marker->markerTransformation(m_markerPositions[i].origin, m_markerPositions[i].angle, strokeWidth);
+            // marker->layer()->paintSVGResourceLayer(paintInfo.context(), LayoutRect::infiniteRect(), contentTransform);
+        }
     }
 }
 
@@ -517,4 +516,27 @@ void RenderSVGShape::processMarkerPositions()
     markerData.pathIsDone();
 }
 
+void RenderSVGShape::styleWillChange(StyleDifference diff, const RenderStyle& newStyle)
+{
+    auto* oldStyle = hasInitializedStyle() ? &style() : nullptr;
+    if (oldStyle) {
+        if (diff == StyleDifference::Layout)
+            setNeedsShapeUpdate();
+    }
+
+    RenderSVGModelObject::styleWillChange(diff, newStyle);
 }
+
+bool RenderSVGShape::needsHasSVGTransformFlags() const
+{
+    return graphicsElement().hasTransformRelatedAttributes();
+}
+
+void RenderSVGShape::applyTransform(TransformationMatrix& transform, const RenderStyle& style, const FloatRect& boundingBox, OptionSet<RenderStyle::TransformOperationOption> options) const
+{
+    applySVGTransform(transform, graphicsElement(), style, boundingBox, std::nullopt, std::nullopt, options);
+}
+
+}
+
+#endif // ENABLE(LAYER_BASED_SVG_ENGINE)
