@@ -28,8 +28,11 @@ class SVGFEFuncGElement final : public SVGComponentTransferFunctionElement {
 public:
     static Ref<SVGFEFuncGElement> create(const QualifiedName&, Document&);
 
+    ComponentTransferChannel channel() const final { return ComponentTransferChannel::Green; }
+
 private:
     SVGFEFuncGElement(const QualifiedName&, Document&);
 };
+static_assert(sizeof(SVGFEFuncGElement) == sizeof(SVGComponentTransferFunctionElement));
 
 } // namespace WebCore
